#ifndef _Rect_h_
#define _Rect_h_

enum EDGE_ID
{
    EDGE_NONE  = 0,
    EDGE_UP    = 1<<0,
    EDGE_RIGHT = 1<<2,
    EDGE_DOWN  = 1<<3,
    EDGE_LEFT  = 1<<4,
		
    EDGE_UL = EDGE_UP | EDGE_LEFT,
    EDGE_UR = EDGE_UP | EDGE_RIGHT,
    EDGE_DL = EDGE_DOWN | EDGE_LEFT,
    EDGE_DR = EDGE_DOWN | EDGE_RIGHT
};

class Rect
{
	protected:
		int x;
		int y;
		int width;
		int height;
		int selected;
	
	public:
		Rect(void);
		Rect(const Rect& r);
		Rect(int nx, int ny, int nw, int nh);
		virtual ~Rect(void);

		virtual void Move(int nx, int ny);
		virtual void MoveTo(const Rect& r);
		virtual void MoveTo(int nx, int ny);
		virtual void Resize(int nw, int nh);
		virtual void ResizeTo(int nw, int nh);
		virtual void ResizeTo(const Rect& r);
		virtual void Change(int nx, int ny, int nw, int nh);
		virtual void ChangeTo(int nx, int ny, int nw, int nh);
		virtual void ChangeTo(const Rect& r);
		
		virtual void Grow(int nx,int ny, int nw, int nh);
		virtual void Grow(const Rect& r);
		
		void Select(void);
		void Deselect(void);
		void InvSelect(void);
		
		int Selected(void) const;
		virtual int InMe(int nx, int ny) const;
		virtual int InMe(const Rect& r) const;
		virtual int OnEgde(int nx, int ny) const;

		int X(void) const;
		int Y(void) const;
		int Width(void) const;
		int Height(void) const;
		Rect Center(void);
		
		virtual void Draw(int mode = 0) const;
};

#endif
