#include <fox/fx.h>
#include <fox/fx3d.h>

#include <GL/gl.h>
#include <GL/glut.h>

#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>
#include <math.h>

#include "resources.h"
#include "VLSIWindow.h"
#include "GroupNode.h"
#include "Layer.h"
#include "Box.h"
#include "Opt.h"

const FXuint TIMER_INTERVAL = 5000;

const FXchar *msk_patterns[]={
	"Layout Files", "*.msk",
	"Text Files", "*.txt",
	"All Files","*.*",
	NULL
};

const FXchar *rul_patterns[]={
	"Rule Files", "*.rul",
	"Text Files", "*.txt",
	"All Files","*.*",
	NULL
};

FXDEFMAP(VLSIWindow) VLSIWindowMap[]=
{
	FXMAPFUNC(SEL_COMMAND,    VLSIWindow::ID_ABOUT,         VLSIWindow::onCmdAbout),
	FXMAPFUNC(SEL_COMMAND,    VLSIWindow::ID_LAYERDLG,      VLSIWindow::onCmdLayerDialog),
	FXMAPFUNC(SEL_UPDATE,     VLSIWindow::ID_LAYERDLG,      VLSIWindow::onUpdLayerDialog),
	
  FXMAPFUNC(SEL_UPDATE,     VLSIWindow::ID_QUERY_MODE,    VLSIWindow::onUpdMode),
	FXMAPFUNC(SEL_TIMEOUT,    VLSIWindow::ID_TIMER,         VLSIWindow::onTimer),
	 
	FXMAPFUNC(SEL_COMMAND,    VLSIWindow::ID_F_NEW,         VLSIWindow::onCmdFileNew),
	FXMAPFUNC(SEL_COMMAND,    VLSIWindow::ID_F_OPEN,        VLSIWindow::onCmdFileOpen),
	FXMAPFUNC(SEL_COMMAND,    VLSIWindow::ID_F_SAVE,        VLSIWindow::onCmdFileSave),
	FXMAPFUNC(SEL_UPDATE,     VLSIWindow::ID_F_SAVE,        VLSIWindow::onUpdFileSave),
	FXMAPFUNC(SEL_COMMAND,    VLSIWindow::ID_F_SAVEAS,      VLSIWindow::onCmdFileSaveAs),
	FXMAPFUNC(SEL_COMMAND,    VLSIWindow::ID_F_QUIT,        VLSIWindow::onCmdFileQuit),

	FXMAPFUNC(SEL_COMMAND,    VLSIWindow::ID_DELETE,        VLSIWindow::onCmdDelete),
	
	FXMAPFUNC(SEL_COMMAND,    VLSIWindow::ID_SELECT_ALL,    VLSIWindow::onCmdSelectAll),
	FXMAPFUNC(SEL_COMMAND,    VLSIWindow::ID_SELECT_NONE,   VLSIWindow::onCmdSelectNone),
	FXMAPFUNC(SEL_COMMAND,    VLSIWindow::ID_SELECT_INVERT, VLSIWindow::onCmdSelectInvert),
	
	FXMAPFUNC(SEL_COMMAND,    VLSIWindow::ID_ZOOMPLUS,      VLSIWindow::onCmdZoomPlus),
	FXMAPFUNC(SEL_COMMAND,    VLSIWindow::ID_ZOOMMINUS,     VLSIWindow::onCmdZoomMinus),
	FXMAPFUNC(SEL_COMMAND,    VLSIWindow::ID_ZOOMALL,       VLSIWindow::onCmdZoomAll),

	FXMAPFUNCS(SEL_COMMAND,   VLSIWindow::ID_MOVEUP, VLSIWindow::ID_MOVERIGHT, VLSIWindow::onCmdMove),
	
	FXMAPFUNC(SEL_COMMAND,    VLSIWindow::ID_SOLOMODE,      VLSIWindow::onCmdSoloMode),
	FXMAPFUNC(SEL_UPDATE,     VLSIWindow::ID_SOLOMODE,      VLSIWindow::onUpdSoloMode),

	FXMAPFUNC(SEL_CONFIGURE,          VLSIWindow::ID_GLCANVAS, VLSIWindow::onGLCanvasConfigure),
	FXMAPFUNC(SEL_PAINT,              VLSIWindow::ID_GLCANVAS, VLSIWindow::onGLCanvasExpose),
	FXMAPFUNC(SEL_MOTION,             VLSIWindow::ID_GLCANVAS, VLSIWindow::onGLCanvasMouseMove),
	FXMAPFUNC(SEL_LEFTBUTTONPRESS,    VLSIWindow::ID_GLCANVAS, VLSIWindow::onGLCanvasMouseLBP),
	FXMAPFUNC(SEL_LEFTBUTTONRELEASE,  VLSIWindow::ID_GLCANVAS, VLSIWindow::onGLCanvasMouseLBR),
	FXMAPFUNC(SEL_RIGHTBUTTONPRESS,   VLSIWindow::ID_GLCANVAS, VLSIWindow::onGLCanvasMouseRBP),
	FXMAPFUNC(SEL_RIGHTBUTTONRELEASE, VLSIWindow::ID_GLCANVAS, VLSIWindow::onGLCanvasMouseRBR),

	FXMAPFUNCS(SEL_UPDATE, VLSIWindow::ID_NOTREADY_FIRST, VLSIWindow::ID_NOTREADY_LAST, VLSIWindow::onUpdNotReady),
};

FXIMPLEMENT(VLSIWindow,FXMainWindow,VLSIWindowMap,ARRAYNUMBER(VLSIWindowMap))

// ======= VLSIWindow::VLSIWindow(FXApp* a) : FXMainWindow(a,"VLSI",NULL,NULL,DECOR_ALL,0,0,700,500)
// VLSIWindow Constructor
VLSIWindow::VLSIWindow(FXApp* a) : FXMainWindow(a,"VLSI",NULL,NULL,DECOR_ALL,0,0,700,500)
{
	#include "load_res.h"
	setIcon(peng_ico);
	#include "VLSIWindow_menu.cpp"
	new FXHorizontalSeparator(this,LAYOUT_SIDE_TOP|SEPARATOR_GROOVE|LAYOUT_FILL_X);
	#include "VLSIWindow_toolbar.cpp"
	
	statusbar=new FXStatusbar(this,LAYOUT_SIDE_BOTTOM|LAYOUT_FILL_X|STATUSBAR_WITH_DRAGCORNER);
		#ifdef DEBUG
			local_y_lab = new FXLabel(statusbar," ",NULL,LAYOUT_RIGHT|FRAME_SUNKEN|LAYOUT_FIX_WIDTH,0,0,32);
			local_x_lab = new FXLabel(statusbar," ",NULL,LAYOUT_RIGHT|FRAME_SUNKEN|LAYOUT_FIX_WIDTH,0,0,32);
				new FXLabel(statusbar,"local:",NULL,LAYOUT_RIGHT);
		#endif
		global_y_lab = new FXLabel(statusbar," ",NULL,LAYOUT_RIGHT|FRAME_SUNKEN|LAYOUT_FIX_WIDTH,0,0,32);
		global_x_lab = new FXLabel(statusbar," ",NULL,LAYOUT_RIGHT|FRAME_SUNKEN|LAYOUT_FIX_WIDTH,0,0,32);
			new FXLabel(statusbar,"position:",NULL,LAYOUT_RIGHT);

		width_lab = new FXLabel(statusbar," ",NULL,LAYOUT_RIGHT|FRAME_SUNKEN|LAYOUT_FIX_WIDTH,0,0,32);
	  height_lab = new FXLabel(statusbar," ",NULL,LAYOUT_RIGHT|FRAME_SUNKEN|LAYOUT_FIX_WIDTH,0,0,32);
			new FXLabel(statusbar,"size:",NULL,LAYOUT_RIGHT);

	// Contents frame
	FXHorizontalFrame* frame = new FXHorizontalFrame(this,LAYOUT_SIDE_TOP|LAYOUT_FILL_X|LAYOUT_FILL_Y);

	glvisual = new FXGLVisual(getApp(),VISUAL_DOUBLEBUFFER);
	glcanvas = new FXGLCanvas(frame,glvisual,this,ID_GLCANVAS,LAYOUT_FILL_X|LAYOUT_FILL_Y);

	new FXTooltip(getApp(),0);
	timer = getApp()->addTimeout(TIMER_INTERVAL,this,ID_TIMER);

	statusline = statusbar->getStatusline();
	statusline->setTarget(this);
	statusline->setSelector(ID_QUERY_MODE);
	status_string = "Welcome to VLSI";
	
	layerdlg = new LayerDlg(this);

	// KONIEC TWORZENIA OBIEKTOW

	filename = "untitled.msk";
	scale = 10;
	base_y = base_x = 0;
	local_x = local_y = 1;
	saved = 1;
	new_file = 1;
	solo_mode = 0;
	active_layer = LR_NWELL;

	mouserect.SetMode(MM_NONE);
	
	for(int i = LR_FIRST; i<=LR_LAST; i++)
	{
		LAYER[i].SetColor(FXRGBA(0,255,0,255));
		LAYER[i].SetId(LR_NULL);
		LAYER[i].Enable();
	}

	LAYER[LR_OPTIONS].SetColor(FXRGBA(255,255,0,200));    LAYER[LR_OPTIONS].SetId(LR_OPTIONS);
	LAYER[LR_METAL1].SetColor(FXRGBA(55,55,255,150));     LAYER[LR_METAL1].SetId(LR_METAL1);
	LAYER[LR_METAL2].SetColor(FXRGBA(0,0,255,150));      	LAYER[LR_METAL2].SetId(LR_METAL2); 
	LAYER[LR_NWELL].SetColor(FXRGBA(132,132,0,255));     	LAYER[LR_NWELL].SetId(LR_NWELL);
	LAYER[LR_DIFFN].SetColor(FXRGBA(0,132,0,180));       	LAYER[LR_DIFFN].SetId(LR_DIFFN);
	LAYER[LR_DIFFP].SetColor(FXRGBA(115,82,0,180));      	LAYER[LR_DIFFP].SetId(LR_DIFFP);
	LAYER[LR_POLY].SetColor(FXRGBA(255,0,0,160));        	LAYER[LR_POLY].SetId(LR_POLY);
	LAYER[LR_CONTACT].SetColor(FXRGBA(225,225,225,225));  LAYER[LR_CONTACT].SetId(LR_CONTACT);
	
	// FIXME: jaki toto ma kolor :/
	LAYER[LR_LOCAL].SetColor(FXRGBA(255,0,0,100));       	LAYER[LR_POLY].SetId(LR_LOCAL);
}

// ======================================================================= VLSIWindow::~VLSIWindow() 
// VLSIWindow Destructor
VLSIWindow::~VLSIWindow() 
{ 
	#ifdef DEBUG
		cerr<<"<"<<this<<"> VLSIWindow::~VLSIWindow start"<<endl;
	#endif
	
	// XXX: dlaczego toto sie nie wykonuje ?!?!?
	
	if(timer!=NULL) getApp()->removeTimeout(timer);
	
	#ifdef DEBUG
		cerr<<"<"<<this<<"> VLSIWindow::~VLSIWindow end"<<endl;
	#endif
}

// ======================================================================= void VLSIWindow::create()
// VLSIWindows Create And Initialize
void VLSIWindow::create()
{
	FXMainWindow::create();
	#include "create_res.h"
	show();
}

// ================================ long VLSIWindow::onGLCanvasConfigure(FXObject*,FXSelector,void*)
// ID_GLCANVAS onGLCanvasConfigure
long VLSIWindow::onGLCanvasConfigure(FXObject*,FXSelector,void*)
{
	#ifdef DEBUG
		cerr<<"<"<<this<<"> VLSIWindow::onGLCanvasConfigure(FXObject*,FXSelector,void*)"<<endl;
	#endif

	if(glcanvas->makeCurrent())
	{
		GLint width = glcanvas->getWidth();
		GLint height = glcanvas->getHeight();

		glViewport(0,0,width,height);

		glLineStipple(4,0xAAAA);

		glEnable(GL_DEPTH_TEST);
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		
		// TODO: stworzyc mapy 32x32 bity dla kazdej z warstw
		// glEnable(GL_POLYGON_STIPPLE);

		glShadeModel(GL_FLAT);
		glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);

		glClearColor(0.0, 0.0, 0.0, 0.0);
		glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
	
		gluOrtho2D(base_x, base_x + width/scale, base_y, base_y + height/scale);

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
//	glTranslatef(0.375,0.375,0.0); // polecane w GL-RedBook, dla renderingu 2D

		glcanvas->makeNonCurrent();
	}
	return 1;
}

// =================================== long VLSIWindow::onGLCanvasExpose(FXObject*,FXSelector,void*)
// ID_GLCANVAS onGLCanvasExpose
long VLSIWindow::onGLCanvasExpose(FXObject*,FXSelector,void*)
{
	if(!glcanvas->makeCurrent())
		return 1;
	
	GLint width = glcanvas->getWidth();
	GLint height = glcanvas->getHeight();

	glViewport(0,0,glcanvas->getWidth(),glcanvas->getHeight());

	glClearColor(0.0, 0.0, 0.0, 0.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	gluOrtho2D(base_x, base_x + width/scale, base_y, base_y + height/scale);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glColor4ub(255,0,0,255);
	glPushMatrix();
		glTranslatef(0.0,0.0,0.990);
		glDisable(GL_LINE_STIPPLE);
		glBegin(GL_LINES);
			glVertex2i(base_x, mouserect.MouseY());
			glVertex2i(base_x + int(width/scale), mouserect.MouseY());
			glVertex2i(mouserect.MouseX(), base_y);
			glVertex2i(mouserect.MouseX(), base_y + int(height/scale));
		glEnd();
	glPopMatrix();

	if(mouserect.GetMode() != MM_NONE)
	{
		glColor4ub(0,255,0,255);
		glPushMatrix();
			glTranslatef(0.0,0.0,0.950);
			mouserect.Draw();
		glPopMatrix();
	}
	
	for(int i=LR_FIRST; i<=LR_LAST; i++) 
	{
		glPushMatrix();
			glTranslatef(0.0,0.0, float(i)*0.005);
			LAYER[i].Draw();
		glPopMatrix();
	}

	if(glvisual->isDoubleBuffer())
		glcanvas->swapBuffers();
		
	glcanvas->makeNonCurrent();

	return 1;
}

// ========================================= long VLSIWindow::onCmdAbout(FXObject*,FXSelector,void*)
// ID_ABOUT - Show AboutBox
long VLSIWindow::onCmdAbout(FXObject*,FXSelector,void*)
{
	FXMessageBox::information(this,MBOX_OK,"About",
		"VLSI 0.0.1b\n"
		"Karolina Wojewoda\n"
		"Bartosz Rempuszewski\n"
		"LODZ 2001");
	return 1;
}

// ======================================= long VLSIWindow::onCmdFileNew(FXObject*,FXSelector,void*)
// ID_F_NEW - New File
long VLSIWindow::onCmdFileNew(FXObject*,FXSelector,void*)
{
	if(!saved)
		switch(FXMessageBox::warning(this,MBOX_YES_NO_CANCEL,"Warning!","Save First?"))
		{
			case MBOX_CLICKED_YES: 
				onCmdFileSave(NULL,0,NULL); 
				if(!saved) return 1; // przerwano dialog zapisu
				break;
			case MBOX_CLICKED_NO: break; // po prostu idziemy dalej			
			case MBOX_CLICKED_CANCEL: return 1;

			default: // oooops! to sie nie powinno stac...
				return 1;
		}

	for(int i=LR_FIRST;i<=LR_LAST;i++)
	{
		LAYER[i].Enable();
		LAYER[i].Action(LA_DELETE,LT_ANY);
	}
	
	new_file=1;
	saved=1;
	filename="untitled.msk";
	StatusText("Ready");

	return 1;
}

// ====================================== long VLSIWindow::onCmdFileOpen(FXObject*,FXSelector,void*) 
// ID_F_OPEN - Open File Dialog
long VLSIWindow::onCmdFileOpen(FXObject*,FXSelector,void*) 
{ 
	FXFileDialog loaddialog(this,"Load Layout File");
	loaddialog.setPatternList(msk_patterns);

	if(!saved)
		switch(FXMessageBox::warning(this,MBOX_YES_NO_CANCEL,"Warning!","Save First?"))
		{
			case MBOX_CLICKED_YES: 
				onCmdFileSave(NULL,0,NULL); 
				if(!saved) return 1; // przerwano dialog zapisu
				break;
			case MBOX_CLICKED_NO: break; // po prostu idziemy dalej			
			case MBOX_CLICKED_CANCEL: return 1;

			default: // oooops! to sie nie powinno stac...
				return 1;
		}
	
	if(!loaddialog.execute())
		return 1;

	if(!fxexists(loaddialog.getFilename().text()))
	{
		FXMessageBox::error(this,MBOX_OK,"Error!",FXString("File:\n" + loaddialog.getFilename() +"\n not found!").text());
		return 1;
	}

	for(int i=LR_FIRST;i<=LR_LAST;i++)
	{
		LAYER[i].Enable();
		LAYER[i].Action(LA_DELETE,LT_ANY);
	}	
	
	filename=loaddialog.getFilename();

	ifstream f;
	f.open(filename.text());
		
	if(f.fail())
	{
		FXMessageBox::error(this,MBOX_OK,"Error!",FXString("File:\n"+filename +"\n fstream::open fail!").text());
		return 1;
	}

	char c;
	int xx,yy,ww,hh;
	char* buffer = new char[80]; // FIXME: jaka jest max. dlugosc lini w pliku MSK?
		
	while(!f.eof())
	{
		c=f.get();
		switch(c)
		{
			case 'R': // REC(x,y,w,h,layer[,opt])
			{
				f.getline(buffer,3,'(');
				f>>xx; f.get(); f>>yy; f.get(); f>>ww; f.get(); f>>hh; f.get(); // pobranie wspolrzednych
				buffer[0] = f.get(); // pobranie pierwszego znaku nazwy warstwy
				buffer[1] = f.get(); // pobranie drugiego znaku
				buffer[2] = '\0'; 

				int l=GetIdByName(buffer); 
				switch(l)
				{
					case LR_OPTIONS:
					{
						Opt* o = new Opt();
						int d;
						f.get(); // "zjedzenie" przecinka
						f>>d;
						o->ChangeTo(xx,yy,ww,hh);
						o->SetData(d);
						LAYER[l].Action(LA_ADD,LT_ARG,(void*) o);
						break;
					}
						
					default:
						Box* b = new Box();
						b->ChangeTo(xx,yy,ww,hh);
						LAYER[l].Action(LA_ADD,LT_ARG,(void*) b);
				}
				f.getline(buffer,79,'\n');
				break;
			} // casr 'R':

			case 'B': // BB(x,y,w,h)
			{
				// po co to niby jest?
				f.getline(buffer,79,'\n');
				break;
			}

			default:
				f.getline(buffer,79,'\n');
		} // switch(c)
	} // while(!f.eof())

	delete [] buffer;		
	f.close();
	glcanvas->update();

	onCmdZoomAll(NULL,0,NULL);
	StatusText("active file " + filename);

	return 1; 
}

// ====================================== long VLSIWindow::onCmdFileSave(FXObject*,FXSelector,void*) 
// ID_SAVE - Save File Dialog
long VLSIWindow::onCmdFileSave(FXObject*,FXSelector,void*) 
{ 
	FXFileDialog loaddialog(this,"Save Layout File As");
	
	loaddialog.setFilename(filename);
	loaddialog.setPatternList(msk_patterns);
	
	if(new_file || !fxexists(filename.text()))
		if(loaddialog.execute())
		{
			if(fxexists(loaddialog.getFilename().text()))
				if((FXMessageBox::warning(this,MBOX_YES_NO,"Warning!",FXString("Owerwrite?\n" + loaddialog.getFilename()).text()) != MBOX_CLICKED_YES))
					return 1;
			filename=loaddialog.getFilename();
		}
		else
			return 1;
					
	filename = loaddialog.getFilename();

	ofstream f;
	f.open(filename.text());
		
	if(f.fail())
	{
		FXMessageBox::error(this,MBOX_OK,"Error!","onCmdFileSave\nfstream::open fail!");
		return 1;
	}
		
	Rect bb;
	for(int i=LR_FIRST; i<=LR_LAST; i++)
		bb.Grow(LAYER[i]);

	f<<"BB("<<bb.X();
	f<<","<<bb.Y();
	f<<","<<bb.Width();
	f<<","<<bb.Height()<<")"<<endl;

	for(int i=LR_FIRST; i<=LR_LAST; i++)
		LAYER[i].Save(f);

	f.close();
	new_file=0;
	saved=1;
	glcanvas->update();

	StatusText("saved: "+filename);

	return 1; 
}

// =============================== long VLSIWindow::onUpdFileSave(FXObject* sender,FXSelector,void*) 
// ID_F_SAVE SEL_UPDATE
long VLSIWindow::onUpdFileSave(FXObject* sender,FXSelector,void* ptr) 
{
	if(saved) sender->handle(this,MKUINT(ID_DISABLE,SEL_COMMAND),ptr);
	else      sender->handle(this,MKUINT(ID_ENABLE,SEL_COMMAND),ptr);
	
	return 1;
}

// ==================================== long VLSIWindow::onCmdFileSaveAs(FXObject*,FXSelector,void*)
// ID_F_SAVEAS SEL_COMMAND
long VLSIWindow::onCmdFileSaveAs(FXObject*,FXSelector,void*)
{
	new_file=1;
	onCmdFileSave(NULL,0,NULL);	
	return 1;
}

// ====================================== long VLSIWindow::onCmdFileQuit(FXObject*,FXSelector,void*)
// ID_F_QUIT SEL_COMMAND
long VLSIWindow::onCmdFileQuit(FXObject*,FXSelector,void*)
{
	switch(FXMessageBox::warning(this,MBOX_QUIT_SAVE_CANCEL,"Warning!","Quit without saving?")) 
	{
		case MBOX_CLICKED_QUIT: 
			getApp()->exit(0); // FIXME: czy TAK sie to robi?, czy wyslac FXApp::ID_QUIT do aplikacji?
			break;

		case MBOX_CLICKED_SAVE:
		{
			bool s = saved;
			saved = 0;
			onCmdFileSave(NULL,0,NULL);
			if(saved) 
				getApp()->exit(0);
			else
				saved = s;
			break;
		}

		case MBOX_CLICKED_CANCEL: 
			return 1;
	}

	return 1;
}

// ============================ long VLSIWindow::onGLCanvasMouseMove(FXObject*,FXSelector,void *ptr)
// ID_GLCANVAS onGLCanvasMouseMove
long VLSIWindow::onGLCanvasMouseMove(FXObject*,FXSelector,void *ptr)
{
	FXEvent *ev =(FXEvent*) ptr;

	local_x = ev->win_x;
	local_y = ev->win_y;

	mouserect.Motion(base_x+int(local_x/scale), base_y+int((glcanvas->getHeight()-local_y)/scale));

	#ifdef DEBUG
		local_x_lab->setText(FXStringVal(local_x));
		local_y_lab->setText(FXStringVal(local_y));
	#endif

	global_x_lab->setText(FXStringVal(mouserect.MouseX()));
	global_y_lab->setText(FXStringVal(mouserect.MouseY()));
	width_lab->setText(FXStringVal(mouserect.Width()));
	height_lab->setText(FXStringVal(mouserect.Height()));

	if(mouserect.GetMode() == MM_LOOK)
	{
		base_x += (local_x - ev->last_x > 0) ? +1 : -1;
		base_y += (ev->last_y - local_y > 0) ? +1 : -1;
	}
	
	glcanvas->update();
	return 1;
}

// ============================= long VLSIWindow::onGLCanvasMouseLBP(FXObject*,FXSelector,void* ptr)
// ID_GLCANVAS - Left Button Pressed
long VLSIWindow::onGLCanvasMouseLBP(FXObject*,FXSelector,void* ptr)
{
	#ifdef DEBUG
		cerr<<"ID_GLCANVAS: Left Button Pressed"<<endl;
	#endif

	if(mouserect.GetMode() != MM_NONE) // zmieniamy tryb tylko jesli nic sie nie dzieje!
		return 1;

	FXEvent *ev =(FXEvent*) ptr;

	local_x = ev->win_x;
	local_y = ev->win_y;

	bool ctrl_down = (ev->state & CONTROLMASK) != 0;
	bool shift_down = (ev->state & SHIFTMASK) != 0;
	
	if(ctrl_down || shift_down) mouserect.SetMode(MM_EDIT);
	else                        mouserect.SetMode(MM_DRAW);
	
	switch(mouserect.GetMode()) 
	{
		case MM_EDIT: 
			mouserect.SetMode(MM_SELECT_ADD);
			if(ctrl_down && shift_down) mouserect.SetMode(MM_SELECT_INV);
			else if(shift_down && !ctrl_down) mouserect.SetMode(MM_SELECT_DEL);

			// statusline->setText("Use mouse to select object(s)");
			StatusText("Use mouse to select object(s)");
			break;
		
		case MM_DRAW:
			mouserect.SetMode(MM_DRAWING);
			// statusline->setText("Use mouse to add object(s)");
			StatusText("Use mouse to add object(s)");
			break;
	};
	
	mouserect.LeftDown(base_x+int(local_x/scale), base_y+int((glcanvas->getHeight()-local_y)/scale));

	glcanvas->update();

	return 1;
}

// ============================= long VLSIWindow::onGLCanvasMouseLBR(FXObject*,FXSelector,void* ptr)
// ID_GLCANVAS - Left Button Released
long VLSIWindow::onGLCanvasMouseLBR(FXObject*,FXSelector,void* ptr)
{
	#ifdef DEBUG
		cerr<<"ID_GLCANVAS: Left Button Released"<<endl;
	#endif

	FXEvent *ev =(FXEvent*) ptr;
	local_x = ev->win_x;
	local_y = ev->win_y;
	
	mouserect.LeftUp(base_x+int(local_x/scale), base_y+int((glcanvas->getHeight()-local_y)/scale));

	bool click = ((mouserect.Width()==0) && (mouserect.Height()==0));
	
	int selectmode = LA_SELECT;

	FXString action;
	switch(mouserect.GetMode())
	{
		case MM_SELECT_ADD: selectmode = LA_SELECT; action="selected"; break;
		case MM_SELECT_DEL: selectmode = LA_DESELECT; action="deselected"; break;
		case MM_SELECT_INV: selectmode = LA_INVSELECT; action="selection changed"; break;
	}

	switch(mouserect.GetMode())
	{
		case MM_SELECT_ADD: case MM_SELECT_DEL: case MM_SELECT_INV: // tryb edycji 
		{
			int count=0;
			if(solo_mode)
				count = LAYER[active_layer].Action(selectmode, click ? LT_RECTXYIN : LT_RECTOVER, (void*) &mouserect);
			else
				for(int i=LR_FIRST;i<=LR_LAST;i++)
					count += LAYER[i].Action(selectmode, click ? LT_RECTXYIN : LT_RECTOVER, (void*) &mouserect);
					
			// statusline->setText(FXStringVal(count)+" object(s) "+action);
			StatusText(FXStringVal(count)+" object(s) "+action);
		break;
		} // case MM_SELECT_???

		case MM_DRAWING: // tryb rysowania
		{
			if(!click)
			{
				if((mouserect.Width()>0) && (mouserect.Height()>0))
				{
					if(active_layer == LR_OPTIONS)
					{
						Opt* o = new Opt();
						o->ChangeTo(mouserect);
						LAYER[active_layer].Action(LA_ADD,LT_ARG, (void*) o);
					}
					else
					{
						Box* b = new Box();
						b->ChangeTo(mouserect);
						LAYER[active_layer].Action(LA_ADD,LT_ARG, (void*) b);
					}
					
					saved = 0;
					// statusline->setText("1 object added");
					StatusText(status_string = "1 object added");
				}
			}
			else // jesli kliknieto to zmieniamy zaznaczenie pojedynczgo obiektu
			{
				int count=0;
				if(solo_mode)
					count=LAYER[active_layer].Action(LA_INVSELECT,LT_RECTXYIN,(void*) &mouserect);	
				else
					for(int i=LR_FIRST;i<=LR_LAST;i++)
						count += LAYER[i].Action(LA_INVSELECT,LT_RECTXYIN,(void*) &mouserect);
				statusline->setText(FXStringVal(count)+" object(s) selection changed");
				StatusText(FXStringVal(count)+" object(s) selection changed");
			}
		break;
		} // case MM_DRAWING
	} // switch(mouserect.GetMode())
	
	mouserect.SetMode(MM_NONE);
	glcanvas->update();
	return 1;
}
// ============================= long VLSIWindow::onGLCanvasMouseRBP(FXObject*,FXSelector,void* ptr)
// ID_GLCANVAS - Right Button Pressed
long VLSIWindow::onGLCanvasMouseRBP(FXObject*,FXSelector,void* ptr)
{
	#ifdef DEBUG
		cerr<<"ID_GLCANVAS: Right Button Pressed"<<endl;
	#endif
	FXEvent *ev =(FXEvent*) ptr;

	local_x = ev->win_x;
	local_y = ev->win_y;

	mouserect.RightDown(base_x+int(local_x/scale), base_y+int((glcanvas->getHeight()-local_y)/scale));
	
	glcanvas->update();
	return 1;
}

// ============================= long VLSIWindow::onGLCanvasMouseRBR(FXObject*,FXSelector,void* ptr)
// ID_GLCANVAS - Right Button Released
long VLSIWindow::onGLCanvasMouseRBR(FXObject*,FXSelector,void* ptr)
{
	#ifdef DEBUG
		cerr<<"ID_GLCANVAS: Right Button Released"<<endl;
	#endif
	FXEvent *ev =(FXEvent*) ptr;

	local_x = ev->win_x;
	local_y = ev->win_y;

	mouserect.RightUp(base_x+int(local_x/scale), base_y+int((glcanvas->getHeight()-local_y)/scale));

	glcanvas->update();
	return 1;
}

// ====================================== long VLSIWindow::onCmdZoomPlus(FXObject*,FXSelector,void*)
// ID_ZOOMPLUS
long VLSIWindow::onCmdZoomPlus(FXObject*,FXSelector,void*)
{
	
	if(scale<=40) 
	{
		base_x += int(glcanvas->getWidth()/scale)/4;
		base_y += int(glcanvas->getHeight()/scale)/4;
		scale *= 2.0;
		if(scale>1) scale = floor(scale); // zaokraglenie w dol
	}

	
	#ifdef DEBUG
		cerr<<"scale(+) == "<<scale<<endl;
	#endif
	glcanvas->update();

	return 1;
}

// ===================================== long VLSIWindow::onCmdZoomMinus(FXObject*,FXSelector,void*)
// ID_ZOOMMINUS
long VLSIWindow::onCmdZoomMinus(FXObject*,FXSelector,void*)
{

	if(scale>0.05) 
	{
		scale /= 2.0;
		if(scale>1) scale = floor(scale);
		base_x -= int(glcanvas->getWidth()/scale)/4;
		base_y -= int(glcanvas->getHeight()/scale)/4;
	}
	
	#ifdef DEBUG
		cerr<<"scale(-) == "<<scale<<endl;
	#endif
	glcanvas->update();
	return 1;
}

// ======================================= long VLSIWindow::onCmdZoomAll(FXObject*,FXSelector,void*)
// ID_ZOOMALL
long VLSIWindow::onCmdZoomAll(FXObject*,FXSelector,void*)
{
	Rect bb;
	for(int i=LR_FIRST; i<=LR_LAST; i++)
		bb.Grow(LAYER[i]);
	
	base_x = bb.X();
	base_y = bb.Y();
	
	float w = bb.Width(); 
	float h = bb.Height();

	#ifdef DEBUG
		cerr<<"BB [x,y,w,h] = ["<<base_x<<","<<base_y<<","<<w<<","<<h<<"]"<<endl;
	#endif
		
	if(w==0) w=1;
	if(h==0) h=1;	

	float xscale = float(glcanvas->getWidth()/w);
	float yscale = float(glcanvas->getHeight()/h);

	scale = (xscale<yscale) ? xscale : yscale;

	if(scale<0.05) scale=0.05;
	if(scale>50) scale=50;

	#ifdef DEBUG
		cerr<<"scale(all) == "<<scale<<endl;
	#endif
	glcanvas->update();
	return 1;
}

// ======================================== long VLSIWindow::onCmdDelete(FXObject*,FXSelector,void*)
// ID_DELETE
long VLSIWindow::onCmdDelete(FXObject*,FXSelector,void*)
{
	int count = 0;
	
	if(solo_mode)
		count = LAYER[active_layer].Action(LA_DELETE,LT_SELECTED);
	else
		for(int i=LR_FIRST;i<=LR_LAST;i++)
			count += LAYER[i].Action(LA_DELETE,LT_SELECTED);

	if(count>0) saved=0;
	// statusline->setText(FXStringVal(count) + " object(s) deleted");
	StatusText(FXStringVal(count) + " object(s) deleted");
	glcanvas->update();
	return 1;
}

// =================================== long VLSIWindow::onCmdLayerDialog(FXObject*,FXSelector,void*)
// ID_LAYERDLG SEL_COMMAND
long VLSIWindow::onCmdLayerDialog(FXObject*,FXSelector,void*)
{
	layerdlg->shown() ? layerdlg->hide() : layerdlg->show();
	this->setFocus();
	return 1;
}

// ======================== long VLSIWindow::onUpdLayerDialog(FXObject* sender,FXSelector,void* ptr)
// ID_LAYERDLG SEL_UPDATE
long VLSIWindow::onUpdLayerDialog(FXObject* sender,FXSelector,void* ptr)
{
	FXuint msg = layerdlg->shown() ? ID_CHECK : ID_UNCHECK;
	sender->handle(this,MKUINT(msg,SEL_COMMAND),ptr);
	this->setFocus();
	return 1;
}

// ====================================== long VLSIWindow::onCmdSoloMode(FXObject*,FXSelector,void*)
// ID_SOLOMODE SEL_COMMAND
long VLSIWindow::onCmdSoloMode(FXObject*,FXSelector,void*)
{
	solo_mode = !solo_mode;
	return 1;
}

// =========================== long VLSIWindow::onCmdSoloMode(FXObject* sender,FXSelector,void* ptr)
// ID_SOLOMODE SEL_UPDATE
long VLSIWindow::onUpdSoloMode(FXObject* sender,FXSelector,void* ptr)
{
	FXuint msg = (solo_mode!=0) ? ID_CHECK : ID_UNCHECK;
	sender->handle(this,MKUINT(msg,SEL_COMMAND),ptr);
	return 1;
}

// ===================================== long VLSIWindow::onCmdSelectAll(FXObject*,FXSelector,void*)
//
long VLSIWindow::onCmdSelectAll(FXObject*,FXSelector,void*)
{
	int count = 0;
	if(solo_mode)
		count = LAYER[active_layer].Action(LA_SELECT,LT_ANY);
	else
		for(int i=LR_FIRST;i<=LR_LAST;i++)
			count += LAYER[i].Action(LA_SELECT,LT_ANY);
	
	StatusText(FXStringVal(count) + " object(s) selected");
	glcanvas->update();
	return 1;
}

// ==================================== long VLSIWindow::onCmdSelectNone(FXObject*,FXSelector,void*)
//
long VLSIWindow::onCmdSelectNone(FXObject*,FXSelector,void*)
{
	int count = 0;
	if(solo_mode)
		count = LAYER[active_layer].Action(LA_DESELECT,LT_ANY);
	else
		for(int i=LR_FIRST;i<=LR_LAST;i++)
			count += LAYER[i].Action(LA_DESELECT,LT_ANY);

	StatusText(FXStringVal(count) + " object(s) deselected");
	glcanvas->update();
	return 1;
}

// ================================== long VLSIWindow::onCmdSelectInvert(FXObject*,FXSelector,void*)
//
long VLSIWindow::onCmdSelectInvert(FXObject*,FXSelector,void*)
{
	int count = 0;
	if(solo_mode)
		count = LAYER[active_layer].Action(LA_INVSELECT,LT_ANY);
	else
		for(int i=LR_FIRST;i<=LR_LAST;i++)
			count += LAYER[i].Action(LA_INVSELECT,LT_ANY);

	StatusText(FXStringVal(count) + " object(s) selection changed");
	glcanvas->update();
	return 1;
}

// =================================== long VLSIWindow::onUpdMode(FXObject* sender,FXSelector,void*)
// ID_QUERY_MODE - Update Status Bar
long VLSIWindow::onUpdMode(FXObject* sender,FXSelector,void*)
{
	FXStatusline *statusline=(FXStatusline*)sender;
	statusline->setText(status_string);

	return 1;
}

// ===================================== long VLSIWindow::onTimer(FXObject* sender,FXSelector,void*)
//
long VLSIWindow::onTimer(FXObject*,FXSelector,void*)
{
	status_string = "Ready";
	timer = getApp()->addTimeout(TIMER_INTERVAL,this,ID_TIMER);
	return 1;
}

// ========================================================= void VLSIWindow::StatusText(FXString s)
//
void VLSIWindow::StatusText(FXString s)
{
	if(timer!=NULL) getApp()->removeTimeout(timer);
	status_string = s;
	timer = getApp()->addTimeout(TIMER_INTERVAL,this,ID_TIMER);
}

// =========================== long VLSIWindow::onUpdNotReady(FXObject* sender,FXSelector,void* ptr)
//
long VLSIWindow::onUpdNotReady(FXObject* sender,FXSelector,void* ptr)
{
	sender->handle(this,MKUINT(ID_DISABLE,SEL_COMMAND),ptr);
	return 1;
}

// ===================== long VLSIWindow::onCmdMove(FXObject* sender, FXSelector selector,void* ptr)
//
long VLSIWindow::onCmdMove(FXObject* sender, FXSelector selector,void* ptr)
{
	switch(SELID(selector))
	{
		case ID_MOVEUP:    base_y += int(glcanvas->getHeight()/scale)/3; break;
		case ID_MOVEDOWN:  base_y -= int(glcanvas->getHeight()/scale)/3; break;
		case ID_MOVERIGHT: base_x += int(glcanvas->getWidth()/scale)/3;  break;
		case ID_MOVELEFT:  base_x -= int(glcanvas->getWidth()/scale)/3;  break;
	}
	glcanvas->update();
	return 1;
}
