#ifndef _VLSIWindow_h_
#define _VLSIWindow_h_

#include <fox/fx.h>
#include <fox/fx3d.h>

#include <GL/gl.h>
#include <GL/glut.h>

#include "resources.h"
#include "MouseRect.h"
#include "Layer.h"
#include "LayerDlg.h"

//#include "Grid.h"
//#include "GridDialog.h"	

class VLSIWindow : public FXMainWindow
{
	FXDECLARE(VLSIWindow)
	
	friend class LayerDlg;
	
	private:
		FXTimer* timer;

		// dialogi
		LayerDlg* layerdlg;
					
		// wigety na glownym oknie
		FXHorizontalFrame* toolbar;
		FXMenuBar* menubar;
		FXGLCanvas* glcanvas;
		FXGLVisual* glvisual;

		FXStatusbar*  statusbar;
		FXStatusline* statusline;
		#ifdef DEBUG
			FXLabel* local_x_lab;
			FXLabel* local_y_lab;
		#endif
			FXLabel* global_x_lab;
			FXLabel* global_y_lab;
			FXLabel* width_lab;
			FXLabel* height_lab;

		MouseRect mouserect;
		// Grid* maingrid;

		// zmienne
		FXint local_x, local_y; // lokalne wspolrzedne myszy na "glcanvas"
		FXint base_x, base_y;   // wsp. bazowe do obl. lokalnych/globalnych wzgledem ekranu
		FXfloat scale; 

		FXint active_layer;

		FXbool solo_mode;
		FXbool saved;
		FXbool new_file;

		Layer LAYER[LR_COUNT]; 
		FXString filename; // nazwa edytowanego pliku
		FXString status_string; // tekst w pasku stanu
		
		// ikony i inne "resource"
		#include "declare_res.h"
		
		public:
			enum
			{
				ID_ABOUT=FXMainWindow::ID_LAST,
				ID_LAYERDLG,
				ID_F_NEW, ID_F_OPEN, ID_F_INSERT, ID_F_SAVE, ID_F_SAVEAS, ID_F_FOUNDRY, ID_F_QUIT,
				ID_DELETE,
				ID_SELECT_ALL, ID_SELECT_NONE, ID_SELECT_INVERT,
				ID_ZOOMPLUS, ID_ZOOMMINUS, ID_ZOOMALL,
				ID_MOVEUP,ID_MOVEDOWN,ID_MOVELEFT,ID_MOVERIGHT,
				ID_SOLOMODE,
				ID_GLCANVAS,
				ID_TIMER,
				ID_QUERY_MODE,
				// XXX: not implemented yet!
				ID_NOTREADY_FIRST,
				ID_COPY, ID_CUT, ID_PASTE,
				ID_UNDO, ID_REDO,
				ID_GROUP, ID_UNGROUP,
				ID_NOTREADY_LAST,

				ID_LAST
			};

		// --------------------------------------------------------------------------------------------
		// rozne funkcje
		void StatusText(FXString); // wyswietla tekst na lini statusu i resetuje "timer"
	
			
		// --------------------------------------------------------------------------------------------
		// obsloga zdarzen (events) 
		
		// Dialogi
		long onCmdLayerDialog(FXObject*,FXSelector,void*);
		long onUpdLayerDialog(FXObject*,FXSelector,void*);
		long onCmdAbout(FXObject*,FXSelector,void*);
	
		// Menu & Toolbar 
		long onCmdFileNew(FXObject*,FXSelector,void*);
		long onCmdFileOpen(FXObject*,FXSelector,void*);
		long onCmdFileSave(FXObject*,FXSelector,void*);
		long onUpdFileSave(FXObject*,FXSelector,void*);
		long onCmdFileSaveAs(FXObject*,FXSelector,void*);
		long onCmdFileQuit(FXObject*,FXSelector,void*);
		
		long onCmdDelete(FXObject*,FXSelector,void*);
		
		long onCmdSelectAll(FXObject*,FXSelector,void*);
		long onCmdSelectNone(FXObject*,FXSelector,void*);
		long onCmdSelectInvert(FXObject*,FXSelector,void*);
		
		long onCmdZoomPlus(FXObject*,FXSelector,void*);
		long onCmdZoomMinus(FXObject*,FXSelector,void*);
		long onCmdZoomAll(FXObject*,FXSelector,void*);

		long onCmdMove(FXObject*,FXSelector,void*);

		long onCmdSoloMode(FXObject*,FXSelector,void*);
		long onUpdSoloMode(FXObject*,FXSelector,void*);

		// GLCanvas 
		long onGLCanvasExpose(FXObject*,FXSelector,void*);
		long onGLCanvasConfigure(FXObject*,FXSelector,void*);
		long onGLCanvasMouseMove(FXObject*,FXSelector,void*);
		long onGLCanvasMouseRBP(FXObject*,FXSelector,void*);
		long onGLCanvasMouseRBR(FXObject*,FXSelector,void*);
		long onGLCanvasMouseLBP(FXObject*,FXSelector,void*);
		long onGLCanvasMouseLBR(FXObject*,FXSelector,void*);

		long onTimer(FXObject*,FXSelector,void*);
		long onUpdMode(FXObject*,FXSelector,void*);

		long onUpdNotReady(FXObject*,FXSelector,void*);
		
		public:
			VLSIWindow(FXApp* a);
			virtual void create();
			virtual ~VLSIWindow();
	
		protected:
			VLSIWindow(){}
};

#endif
